/*===========================================================================*/
/*
 * This file is part of
 * libneuro - a C++ library for signals in the Neuro format
 *
 * Copyright (C) 2006, 2007, 2008  Elaine Tsiang YueLien
 *
 * libneuro is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, see http://www.gnu.org/licenses, or write
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301, USA
 *
 *===========================================================================*/
/** @file NeuroImpl.H                                                        *
 *                                                                           *
 *  Neuro::NeuroImpl - Neuro implementation interface                        *
 *                                                                           */
/*===========================================================================*/
#ifndef	NeuroImplementationInterface
#define	NeuroImplementationInterface

#ifdef __GNUG__
#pragma interface
#endif

#include	<config.h>
#include	<ALingA/Encode.H>
#include	<ALingA/XMLwriter.H>
#include	<ALingA/XMLparser.H>
#include	<Neuro/Neuro.H>

namespace	Neuro
{
  using namespace	Ogg;
  using namespace	ALingA;

  class	NeuroImpl;

  class	XMLparser;
  class NeuroXMLparser;
  class FrameTypeXMLparser;


  class NeuroImpl
  {
    friend class	Neuro;
    friend class	iFrames;
    friend class	XMLparser;
    friend class	NeuroXMLparser;
    friend class	FrameTypeXMLparser;

  private:
    class	iFramesImpl
      : public Neuro::iFrames
    {
    public:
      NeuroImpl &		neuro_;
      Logical::Reader *		reader_;
      Logical::Writer *		writer_;
      Signal *			signal_;

      char *			frames_;
      size_t			numFrames_;
      size_t			frameSize_;
      size_t			packetSize_;
      size_t			currentPacketSize_;
      size_t			gransPerFrame_;
      size_t			totalGrans_;

      iFramesImpl(
		  NeuroImpl &
		  ,Signal * signal = 0
		  );

      ~iFramesImpl()
      {}

      ALingA::Position
      basePosition()
      {
	if ( !signal_ )
	  return(totalGrans_ -1);

	return( signal_->basePosition(ALingA::Position(numFrames_-1)
				      ,ALingA::Position(gransPerFrame_-1)
				      )
		);
      }

    };

  private:
    static
    const size_t		minimumPacketSize_ = 1024;
    static
    const unsigned short	dtdMajor_ = NEURO_DTD_MAJOR;
    static
    const unsigned short	dtdMinor_ = NEURO_DTD_MINOR;

    

    Neuro *			intf_;
    bool			writing_;
    Glib::ustring		xmlns_;
    Manifold			manifold_;
    Encode			encode_;
    bool			dataIsBigEndian_;
    size_t			framesPerPacket_;
    bool			doneParsingHeaders_;

    Metas			metas_;

    NeuroImpl(
	      Manifold &
	      ,Encode &
	      ,long		framesPerPacket
	      );
    NeuroImpl();

    ~NeuroImpl()
    {}

    void
    writeHeaders(
		 Logical::Writer &
		 );
	
    void
    readHeaders(
		Logical::Reader &
		);
  };

  class XMLparser : public ALingA::XMLparser
  {
    friend class	NeuroImpl;
    friend class	ALingA::XMLparser;

  private:
    NeuroImpl &		neuroImpl_;

    XMLparser(
	      NeuroImpl &	neuroImpl
	      )
      : ALingA::XMLparser("Driver")
      , neuroImpl_(neuroImpl)
    {}

    ~XMLparser()
    {}

    /* base behaviour
       void
       on_start_document();

       void
       on_end_document();
    */

    void
    on_start_element(
		     const Glib::ustring& 	name
		     , const xmlpp::SaxParser::AttributeList& attributes
		     );
    
    void
    on_end_element(
		   const Glib::ustring& 	name
		   )
    {
      throw XMLparser::UnexpectedElement(name);
    }
    
    void
    on_characters(
		  const Glib::ustring& characters
		  )
    {
      if ( characters != "\n" )
	throw ALingA::XMLparser::UnexpectedCharacters(characters);

    }
    
    void on_cdata_block(
			const Glib::ustring& text
			)
    {
      throw ALingA::XMLparser::UnexpectedCData(text);
    }
    
  };

  class NeuroXMLparser : public ALingA::XMLparser
  {
    friend class	NeuroImpl;
    friend class	XMLparser;

  private:
    NeuroImpl &		neuroImpl_;

  public:
    NeuroXMLparser(
		   XMLparser &			xmlParser
		   , NeuroImpl &		neuroImpl
		   , const xmlpp::SaxParser::AttributeList& attributes
		   )
      ;

    ~NeuroXMLparser()
    {}


    void
    on_start_document()
    {
      throw ALingA::XMLparser::UnexpectedDocument();
    }
    

    void
    on_end_document()
    {
      throw ALingA::XMLparser::UnexpectedDocument();
    }
    


    void
    on_start_element(
		     const Glib::ustring& 	name
		     , const xmlpp::SaxParser::AttributeList& attributes
		     );

    void
    on_end_element(
		   const Glib::ustring& 	name
		   );

    void
    on_characters(
		  const Glib::ustring& characters
		  )
    {
      if ( characters != "\n" )
	throw ALingA::XMLparser::UnexpectedCharacters(characters);
    }
    
    void on_cdata_block(
			const Glib::ustring& text
			)
    {
      throw ALingA::XMLparser::UnexpectedCData(text);
    }
  };

  class FrameTypeXMLparser : public ALingA::XMLparser
  {
  private:
    NeuroImpl &		neuroImpl_;

  public:
    FrameTypeXMLparser(
		       XMLparser &		neuroXmlParser
		       ,NeuroImpl &		neuroImpl
		       , const xmlpp::SaxParser::AttributeList& attributes
		       )
      ;

    ~FrameTypeXMLparser()
    {}


    void
    on_start_document()
    {
      throw ALingA::XMLparser::UnexpectedDocument();
    }
    

    void
    on_end_document()
    {
      throw ALingA::XMLparser::UnexpectedDocument();
    }
    


    void
    on_start_element(
		     const Glib::ustring& 	name
		     , const xmlpp::SaxParser::AttributeList& attributes
		     );
    /* base behaviour
       void
       on_end_element(
       const Glib::ustring& 	name
       );
    */

    
    void
    on_characters(
		  const Glib::ustring& characters
		  )
    {
      if ( characters != "\n" )
	throw ALingA::XMLparser::UnexpectedCharacters(characters);
    }
    
    void on_cdata_block(
			const Glib::ustring& text
			)
    {
      throw ALingA::XMLparser::UnexpectedCData(text);
    }
  };

}

#endif
