/*===========================================================================*/
/*
 * This file is part of
 * libneuro - a C++ library for signals in the Neuro format
 *
 * Copyright (C) 2006, 2007, 2008  Elaine Tsiang YueLien
 *
 * libneuro is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, see http://www.gnu.org/licenses, or write
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301, USA
 *
 *===========================================================================*/
/** @file raw2Neuro.C                                                        */
/*                                                                           */
/* raw2Neuro - convert raw signal file to Neuro fromat                       */
/*                                                                           */
/*===========================================================================*/
#include	<Neuro/Neuro.H>
#include	<Thread.H>
#include	<Ogg/Debug.H>

#include	<fstream>

extern "C" 
{
#include	<string.h>
#include	<unistd.h>
}

namespace	Examples
{
  using namespace	Ogg;
  using namespace	ALingA;
  using namespace	Neuro;

  class Muxer: public Transport, public Thread
  {
    std::ofstream	f;

  public:
    Muxer()
      : Transport(true)
    {}

    void
    run()
    {
      loop();
    }

    void
    sendCallback(
		 Page &	pg
		 )
    {
      if ( std::cout.write(static_cast<char *>(pg.data())
			   , pg.size()
			   )
	   )
	{
	  debug(true,
		"Serial No. ", pg.serialNo()
		," sent page ",pg.pageNo()
		," with granule position ", pg.granulePosition()
		);
	}
      else
	{
	  debug(true,
		"Serial No. ", pg.serialNo()
		," error occured sending page ",pg.pageNo()
		," with granule position ", pg.granulePosition()
		);
	}

      if ( pg.ending() )
	{
	  debug(true
		,"Serial No. ", pg.serialNo()
		," wrote ending page ",pg.pageNo()
		);
	  terminate();
	}
    }

  }
  ;

  class NeuroWriter : public ::Neuro::Neuro, public Thread
  {
  public:
    NeuroWriter(
		Muxer &		muxer
		,long		track
		,Manifold &	manifold
		,Encoding	encoding
		,long		framesPerPacket
		)
      : Neuro(muxer
	      ,track
	      ,manifold
	      ,encoding
	      ,framesPerPacket
	      )
    {}

    void
    run()
    {
      try
	{
	  iFrames & i = NeuroWriter::write();

	  do 
	    {
	      if ( std::cin.read(reinterpret_cast<char *>(i.frame())
				 ,i.frameSize()
				 )
		   ||
		   std::cin.eof()
		   )
		{
		  if ( std::cin.gcount() == i.frameSize() )
		    {
		      if ( 0 == i.frameNo() % framesPerPacket() )
			debug(true,
			      "got ",i.frameNo(),"th frame"
			      );
		      ++i;
		    }
		  else
		    {
		      // incomplete frame
		      // zero unfilled
		      for ( long j = std::cin.gcount()
			      ;
			    (j< i.frameSize())
			      ;
			    ++j
			    )
			{
			  reinterpret_cast<char *>(i.frame())[j] = 0;
			}
		    }
		}
	    }
	  while ( !std::cin.eof() );
	  
	  debug(true,
		"got last ",i.frameNo(),"th frame"
		);

	  // destroy i here to write out last frame
	  delete &i;

	}
      catch ( exception ex
	      )
	{
	  debug(true,
		ex.what()
		);
	  throw;
	}

      debug(true
	    ,"exiting NeuroWriter::run()"
	    );
      return;
    }
      
  };

}

int main(int argc, const char* argv[])
{
  using namespace	Examples;
  using namespace	Ogg;

  Muxer & muxer	= *(new Muxer);
  Manifold	manifold(16000);		// sampled at 16KHz 
  Encoding	encoding = Signed16Bit;
  long		framesPerPacket(1024);

  NeuroWriter & neuroWriter = *(new NeuroWriter(muxer
						,1
						,manifold
						,encoding
						,framesPerPacket
						));
  // host may not be, but data is already BigEndian
  // neuroWriter.dataIsBigEndian(true);


  muxer.start();
  neuroWriter.start();

  neuroWriter.finish();
  muxer.finish();

  exit(0);
}
